package ch.hepia.numeric;
import java.util.List;
import java.util.ArrayList;
import java.util.function.BiFunction;
import java.util.function.DoubleFunction;
import ch.hepia.numeric.Vector;
public class Matrix {
    private List<Vector> col;

    private Matrix(List<Vector> vecLst){
        this.col = vecLst;
    }
    public int nbRows() {
        return this.col.get(0).len();
    }
    public int nbCols() {
        return this.cols().size();
    }
    public List<Vector> cols() {
        return this.col;
    }
    public List<Transposed> rows(){
        List<Transposed> lstTran = new ArrayList<>();
        for(int i = 0;i < this.nbRows();i++){
            List<Double> newTran = new ArrayList<>();
            for (Vector x : this.cols()){
                newTran.add(x.get(i));
            }
            lstTran.add(new Transposed(newTran));
        }
        return lstTran;
    }
    public Transposed getRow(int i) {
        return this.rows().get(i);
    }
    public Vector getCol(int i) {
        return cols().get(i);
    }
    public double get(int i, int j) {
        return this.cols().get(i).get(j);
    }
    void set(int row, int col, double value) {
        this.col.get(col).set(row, value);
    }
    public Matrix map(DoubleFunction<Double> f) {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public Matrix add(Matrix that) {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public Matrix mul(double d) {
        List<Vector> lstVec = new ArrayList<>();
        for (int i = 0; i < this.cols().size();i++){
            lstVec.add(this.cols().get(i).mul(d));
        }
        return Matrix.of(lstVec);
    }
    public Matrix mul(Matrix that) {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public Matrix sub(Matrix that) {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public boolean isSquare() {
        if (this.nbCols() == this.nbRows()) 
        {
            return true;
        }
        return false;
    }
    public Matrix t() {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public Matrix rowRemoved(int i){
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public Matrix colRemoved(int i){
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public Matrix removed(int i, int j) {
        List<Vector> lstVec = new ArrayList<>();
        for (int x = 0; x < this.nbCols(); x++){
            List<Double> newVec = new ArrayList<>();
            for (int y = 0; y < this.nbRows(); y++){
                if ( x != i &&  y != j){
                    newVec.add(this.get(x, y));
                }
            }
            if (newVec.size() > 0){
                lstVec.add(Vector.of(newVec));
            }
        }
        return Matrix.of(lstVec);
    }

    public Matrix adjugate() {
        Matrix res = Matrix.tabulate(nbRows(), nbCols(), (i, j) -> {
            if ((i + j) % 2 == 0) {
                return removed(i, j).det();
            } else {
                return -removed(i, j).det();
            }
        });
        return res;
    }

    public Matrix inv() {
        Matrix newMat = this.adjugate();
        newMat = newMat.mul(1.0/this.det());
        return newMat;

    }

    public class notASquareException extends Exception{
        public notASquareException(){
            super("The matrix is not a square ! ");
        }
    }

    public double det() {
        if (!isSquare()) {
            //throw new notASquareException();
        }

        if( nbCols() == 1) {
            return get(0,0);
        } else if( nbCols() == 2 ){
            return get(0,0)*get(1,1) - get(1,0)*get(0,1);
        } else {
            double tot = 0.0;
            for (int i = 0; i < this.nbCols();i++){
                if (i % 2 == 0){
                    tot += this.removed(i, 0).det() * this.get(i,0);
                }
                else{
                    tot += this.removed(i, 0).det() * this.get(i,0) * (-1.0);
                }
            }
            return tot;
        }
    }

    public Matrix copy() {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }

    @Override
    public String toString() {
        String str = "";
        for (Vector x : this.cols()){
            str += "\n" + x.toString();
        }
        return str;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) return true;
        if (obj == null || obj.getClass() != this.getClass()) {
            return false;
        }
        Matrix mat = (Matrix) obj;
        if (mat.nbRows() != this.nbRows() || mat.nbCols() != this.nbCols()){
            return false;
        }
        for (int i = 0; i < this.nbCols();i++){
            for (int j = 0; j < this.nbRows();j++){
                if (this.get(i, j) != mat.get(i, j)){
                    return false;
                }
            }
        }
        return true;
    }

    public static Matrix of(List<Vector> ts) {
        return new Matrix(ts);
    }
    public static Matrix of(Vector... ts) {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public static Matrix empty() {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public static Matrix fill(int rows, int cols, double value) {
        List<Vector> newLst = new ArrayList<>();
        for (int i = 0; i < cols; i++){
            newLst.add(Vector.fill(rows, value));
        }
        return Matrix.of(newLst);
    }
    public static Matrix zeros(int rows, int cols) {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public static Matrix ones(int rows, int cols) {
        throw new UnsupportedOperationException("This feature isn't implemented yet"); 
    }
    public static Matrix tabulate(int rows, int cols, BiFunction<Integer, Integer, Double> f) {
        Matrix newMat = Matrix.fill(rows, cols, 0);
        for (int i=0; i < newMat.nbCols();i++){
            for (int y=0; y < newMat.nbRows();y++){
                newMat.set(y, i, f.apply(y, i));
            }
        }
        return newMat;
    }
    public static Matrix diag(double... ds) {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public static Matrix identity(int nb) {
        throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    public static Vector solve(Matrix m, Vector b) {
        Matrix newMat = m.inv();
        List<Double> newLst = new ArrayList<>();
        for (int i = 0; i < newMat.nbCols(); i++){
            double tot = 0.0;
            for (int y = 0; y < b.len(); y++){
                tot += newMat.cols().get(y).get(i) * b.get(i);
            }
            newLst.add(tot);
        }
        return Vector.of(newLst);
    }
}










